/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/

#include "ironman.h"

#ifndef _GRID_H
#define _GRID_H

#include "struct.h"
#include "ironman.h"
#include "zplglobal.h"
#include "type.h"
#include "direction.h"

#define _UNIPROC(numdims) ((1 << (numdims)) - 1)

#define PROCESSORS (_PROCESSORS)
#define GRIDPROCS(grd, dim) (_GRD_SIZE(grd, dim - 1))
#define GRIDPROC(grd, dim) (_GRD_LOC(grd, dim - 1))

#define _NEXTPROC ((_INDEX+1)%_PROCESSORS)
#define _LEGALPROC (_INDEX < _PROCESSORS)

#define _GRD_SIZE_V(grd) ((grd).size)
#define _GRD_SIZE(grd, dim) ((grd).size[(dim)])
#define _GRD_LOC_V(grd) ((grd).loc)
#define _GRD_LOC(grd, dim) ((grd).loc[(dim)])
#define _GRD_BLK_V(grd) ((grd).blocksize)
#define _GRD_BLK(grd, dim) ((grd).blocksize[(dim)])
#define _GRD_NUMSLICES(grd) ((grd).numslices)
#define _GRD_DISTRIBUTED_V(grd) ((grd).distributed)
#define _GRD_DISTRIBUTED(grd, slc) ((grd).distributed[slc])

#define _GRD_LABEL(grd) ((grd).grid_label)

#define _GRD_BEG(grd, dim) (_GRD_LOC(grd, dim) == 0)
#define _GRD_END(grd, dim) (_GRD_LOC(grd, dim) == _GRD_SIZE(grd, dim) - 1)

#define _GRD_NUM_USEFUL_COMM_DIRS(grd) ((grd).num_useful_comm_dirs)
#define _GRD_NUM_USEFUL_AT_DIRS(grd) ((grd).num_useful_at_dirs)
#define _GRD_DIRPTR(grd) ((grd).dirptr)

#define _GRD_REGISTER(grd, dom) _grid_register_domain(grd, dom)
#define _GRD_UNREGISTER(grd, dom) _grid_unregister_domain(grd, dom)
#define _GRD_DOMLIST(grd) ((grd).domains)
#define _GRD_DOMLIST_DOM(gdlist) ((gdlist)->dom)
#define _GRD_DOMLIST_NEXT(gdlist) ((gdlist)->next)

typedef struct {
  _region_pnc reg;
  _array_pnc arr;
  _vector lo;
  _vector hi;
  unsigned int elemsize;
  char* buffer;
} spsinfo;

typedef struct {
  int numsps;
  spsinfo* vect;
} spsinfovect;

typedef struct {
  spsinfovect recvsps;
  spsinfovect sendsps;
} spscomminfo;

typedef struct {
  int recvready;
  IM_comminfo* recvinfo;
  int sendready;
  IM_comminfo* sendinfo;
} comminfo;

typedef struct {
  unsigned int elemsize;
  _array_pnc arr;
  int lhscomm;
  int numwrapdirs;
  int numdirs;
  _direction_pnc* dirlist;
} _arrcomminfo;

typedef struct {
  int numarr;
  _arrcomminfo* arrlist;
} _atcomminfo;

typedef struct {
  int remcomms;
  int lclcomms;
} commcount;

struct __domain_list {
  _domain_nc dom;
  _domain_list* next;
};

struct __grid_info {
  int processors;          /* number of processors involved            */
  int index;               /* my index number                          */
  int size[_MAXRANK];      /* number of processors per dimension       */
  int loc[_MAXRANK];       /* my location in the processor grid        */
  int blocksize[_MAXRANK]; /* multiplicative factors for conversion of
	 		      logical indices to/from processor index  */
  int numslices;           /* number of slices = 2^rank                */
  int* distributed;        /* 1 if slice is distributed; 0 otherwise   */

  int num_useful_comm_dirs;     /* at_comm stuff */
  int num_useful_at_dirs;       /* at_comm stuff */
  int* dirptr;                  /* at_comm stuff */

  _domain_list* domains;
  char* grid_label;
  _MD_GRID_INFO            /* machine dependent info, e.g. proc groups */
};

_grid_info _DefaultGrid;

int _PROCESSORS;
int _INDEX;

void _grid_register_domain(_grid_fnc, _domain_fnc);
void _grid_unregister_domain(_grid_fnc, _domain);
void _print_grid_hierarchy(_grid);

#endif
