/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/

#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <float.h>   /* float needs to come after limits.h on dept sgis */
#include "md_zinc.h"
#include "priv_access.h"
#include "region.h"
#include "grid.h"
#include "domain.h"
#include "zplglobal.h"
#include "zerror.h"
#include "zlib.h"
#include "zmacros.h"

_bound _invisibleBounds[_MAXRANK];

static void _CalcRegionBounds(_region_fnc reg, int dim) {
  _domain dom = _REG_DOM(reg);
  _grid grd = _DOM_GRD(*dom);

  if (dim >= _NUMDIMS(reg)) {
    _MIN_PROC(reg, dim) = 0;
    _MAX_PROC(reg, dim) = 0;
  } else {
    _MIN_PROC(reg, dim) = _DOM_TO_PROC(unknown, *dom, _REG_GLOB_LO(reg, dim), dim);
    if (_MIN_PROC(reg, dim) >= _GRD_SIZE(*grd, dim)) {
      _MIN_PROC(reg, dim) = _GRD_SIZE(*grd, dim)-1;
      _MAX_PROC(reg, dim) = _GRD_SIZE(*grd, dim)-1;
      return;
    } else if (_MIN_PROC(reg, dim) < 0) {
      _MIN_PROC(reg, dim) = 0;
    }
    _MAX_PROC(reg, dim) = _DOM_TO_PROC(unknown, *dom, _REG_GLOB_HI(reg, dim), dim);
    if (_MAX_PROC(reg, dim) < 0) {
      _MAX_PROC(reg, dim)=0;
    } else if (_MAX_PROC(reg, dim) >= _GRD_SIZE(*grd, dim)) {
      _MAX_PROC(reg, dim)=_GRD_SIZE(*grd, dim) - 1;
    }
  }
}

static void _SetupRegionDistribution(_region_fnc reg) {
  _domain dom = _REG_DOM(reg);
  int i;
  int hi;
  _PRIV_TID_DECL;

  for (i=0;i<_NUMDIMS(reg);i++) {
    if (_FLOODED_DIM(reg,i)) {
      _SET_GLOB_BOUNDS_FLOOD(reg, i, _DOM_LO(*dom, i), _DOM_HI(*dom, i), _FLOODED_DIM(reg,i));
    }
  }
  hi = _MAXRANK;
  for (i=0;i<hi;i++) {
    _CalcRegionBounds(reg,i);
  }
}

void _SetMyBlock(int dim, _region_fnc reg, int domlo, int domhi, _grid grd) {
  const int lo = _REG_GLOB_LO(reg, dim);
  const int hi = _REG_GLOB_HI(reg, dim);
  int mylo;
  int myhi;

  if ((lo > domhi && !_GRD_END(*grd, dim)) ||
      ((hi < domlo) && !_GRD_BEG(*grd, dim))) {
    mylo = 0;
    myhi = -1;
  } else {
    if (lo < domlo && !_GRD_BEG(*grd, dim)) {
      mylo = _SNAP_UP(domlo, reg, dim);
    } else {
      mylo = lo;
    }

    if (hi > domhi && !_GRD_END(*grd, dim)) {
      myhi = _SNAP_DN(domhi, reg, dim);
    } else {
      myhi=hi;
    }

    if (mylo > myhi) {
      mylo = 0;
      myhi = -1;
    }
  }
  _REG_MYLO(reg, dim) = mylo;
  _REG_MYHI(reg, dim) = myhi;
}


static void _GrabDimension(_region_fnc reg,int dim) {
  _PRIV_TID_DECL;

  _REG_MYLO(reg, dim) = _REG_GLOB_LO(reg, dim);
  _REG_MYHI(reg, dim) = _REG_GLOB_HI(reg, dim);
}


static void _DistribDimension(_region_fnc reg,int dim) {
  _domain dom = _REG_DOM(reg);
  _grid grd = _DOM_GRD(*dom);
  _PRIV_TID_DECL;

  _SetMyBlock(dim, reg, _DOM_MYLO(*dom, dim), _DOM_MYHI(*dom, dim), grd);
}


static void _SetupGeneralRegion(_region_fnc reg) {
  _domain dom = _REG_DOM(reg);
  _grid grd = _DOM_GRD(*dom);
  int numdims;
  int i;
  int igetnothing;

  numdims = _NUMDIMS(reg);
  for (i = 0; i < numdims; i++) {
    if (_FLOODED_DIM(reg, i)) {
      _REG_MYLO(reg, i) = _DOM_MYLO(*dom, i);
      _REG_MYHI(reg, i) = _DOM_MYLO(*dom, i);
      _REG_STRIDE(reg, i) = 1;
    } else if (_MIN_PROC(reg, i) == _MAX_PROC(reg, i)) {
      if (_MIN_PROC(reg, i) == _GRD_LOC(*grd, i)) {
	_GrabDimension(reg, i);
      } else {
	_REG_MYLO(reg, i) = 0;
	_REG_MYHI(reg, i) = -1;
      }
    } else {
      _DistribDimension(reg, i);
    }
  }
  if (numdims < _MAXDIM) {
    igetnothing = 0;
    for (i = numdims; i < _MAXDIM; i++) {
      _REG_MYLO(reg, i) = 0;
      _REG_MYHI(reg, i) = -1;
      if (!_GRD_BEG(*grd, i)) {
	igetnothing = 1;
      }
    }
    if (igetnothing) {
      for (i = 0; i < numdims; i++) {
	_REG_MYLO(reg, i) = 0;
	_REG_MYHI(reg, i) = -1;
      }
    }
  }
  _REG_I_OWN(reg) = 1;
  for (i = 0; i < numdims; i++) {
    if (_REG_MYLO(reg, i) > _REG_MYHI(reg, i)) {
      _REG_I_OWN(reg) = 0;
    }
  }
}


void _SetupRegion(_region_fnc reg) {
  _SetupRegionDistribution(reg);
  _SetupGeneralRegion(reg);
}

void _InitDomain(_domain_fnc dom, _grid_fnc grd, char* label) {
  _DOM_GRD(*dom) = grd;
  /*  _GRD_REGISTER(grd, dom);*/
  _DOM_LABEL(*dom) = label;
}

void _domain_register_region(_domain_fnc dom, _region_fnc reg) {
  _region_list* tmp = _zmalloc(sizeof(_region_list), "domain region list node");

  _DOM_REGLIST_REG(tmp) = reg;
  _DOM_REGLIST_NEXT(tmp) = _DOM_REGLIST(*dom);
  _DOM_REGLIST(*dom) = tmp;
}

void _domain_unregister_region(_domain_fnc dom, _region reg) {
  _region_list* tmp;
  _region_list* prev;

  prev = NULL;
  tmp = _DOM_REGLIST(*dom);
  while (tmp) {
    if (_DOM_REGLIST_REG(tmp) == reg) {
      break;
    }
    prev = tmp;
    tmp = _DOM_REGLIST_NEXT(tmp);
  }
  if (!tmp) {
    _RT_ANY_FATAL1("Region %s cannot be unregistered", _REG_LABEL(reg));
  }
  if (prev) {
    _DOM_REGLIST_NEXT(prev) = _DOM_REGLIST_NEXT(tmp);
  }
  else {
    _DOM_REGLIST(*dom) = _DOM_REGLIST_NEXT(tmp);
  }
  _zfree(tmp, "domain region list node");
}

void _print_domain_hierarchy(_domain dom) {
  _region_list* tmp = _DOM_REGLIST(*dom);

  _SCALAR_FPRINTF1(zout, "Domain %s\n", _DOM_LABEL(*dom));
  while (tmp) {
    _print_region_hierarchy(_DOM_REGLIST_REG(tmp));
    tmp = _DOM_REGLIST_NEXT(tmp);
  }
}


void _blockedSetup(_domain_fnc dom, int dim, int lo, int hi) {
  _grid grd = _DOM_GRD(*dom);
  double bigval;

  if (lo == INT_MAX && hi == INT_MIN) {
    lo = 1;
    hi = _GRD_SIZE(*grd, dim);
  }
  if (_GRD_SIZE(*grd, dim) > hi - lo + 1) {
    hi = lo + _GRD_SIZE(*grd, dim) - 1;
  }
  bigval = (double)(hi - lo + 1) * _GRD_LOC(*grd, dim);
  _DOM_MYLO(*dom, dim) = _PROC_TO_DATA((bigval) / _GRD_SIZE(*grd, dim)) + lo;
  bigval = (double)(hi - lo + 1) * (_GRD_LOC(*grd, dim) + 1);
  _DOM_MYHI(*dom, dim) = _PROC_TO_DATA((bigval) / _GRD_SIZE(*grd, dim)) - 1 + lo;
  _DOM_LO(*dom, dim) = lo;
  _DOM_HI(*dom, dim) = hi;
  _DIS_TO_PROC(*_DOM_DIS(*dom, dim)) = _blockedToProc;
}

int _blockedToProc(_domain dom, int dim, int i) {
  return ((int)((((double)(i) - _DOM_LO(*dom, dim)) * _DOM_GRDSIZE(*dom, dim)) / _DOM_SIZE(*dom, dim)));
}

void _gooferSetup(_domain_fnc dom, int dim) {
  _grid grd = _DOM_GRD(*dom);

  if (_GRD_SIZE(*grd, dim) == 1) {
    _DOM_MYLO(*dom, dim) = INT_MIN;
    _DOM_MYHI(*dom, dim) = INT_MAX;
  }
  else if (_GRD_LOC(*grd, dim) == 0) {
    _DOM_MYLO(*dom, dim) = INT_MIN;
    _DOM_MYHI(*dom, dim) = 0;
  }
  else if (_GRD_LOC(*grd, dim) == _GRD_SIZE(*grd, dim) - 1) {
    _DOM_MYLO(*dom, dim) = _GRD_LOC(*grd, dim);
    _DOM_MYHI(*dom, dim) = INT_MAX;
  }
  else {
    _DOM_MYLO(*dom, dim) = _GRD_LOC(*grd, dim);
    _DOM_MYHI(*dom, dim) = _GRD_LOC(*grd, dim);
  }
  _DOM_LO(*dom, dim) = INT_MIN;
  _DOM_HI(*dom, dim) = INT_MAX;
  _DIS_TO_PROC(*_DOM_DIS(*dom, dim)) = _gooferToProc;
}

int _gooferToProc(_domain dom, int dim, int i) {
  _grid grd = _DOM_GRD(*dom);

  if (i <= 0) {
    return 0;
  }
  else if (i >= _GRD_SIZE(*grd, dim)-1) {
    return _GRD_SIZE(*grd, dim)-1;
  }
  else {
    return i;
  }
}

void _fourferSetup(_domain_fnc dom, int dim) {
  _grid grd = _DOM_GRD(*dom);

  if (_GRD_SIZE(*grd, dim) == 1) {
    _DOM_MYLO(*dom, dim) = INT_MIN/2;
    _DOM_MYHI(*dom, dim) = INT_MAX/2;
  }
  else if (_GRD_LOC(*grd, dim) == 0) {
    _DOM_MYLO(*dom, dim) = INT_MIN/2;
    _DOM_MYHI(*dom, dim) = 3;
  }
  else if (_GRD_LOC(*grd, dim) == _GRD_SIZE(*grd, dim) - 1) {
    _DOM_MYLO(*dom, dim) = _GRD_LOC(*grd, dim) * 4;
    _DOM_MYHI(*dom, dim) = INT_MAX/2;
  }
  else {
    _DOM_MYLO(*dom, dim) = _GRD_LOC(*grd, dim) * 4;
    _DOM_MYHI(*dom, dim) = _GRD_LOC(*grd, dim) * 4 + 3;
  }
  _DOM_LO(*dom, dim) = INT_MIN/2;
  _DOM_HI(*dom, dim) = INT_MAX/2;
  _DIS_TO_PROC(*_DOM_DIS(*dom, dim)) = _fourferToProc;
}

int _fourferToProc(_domain dom, int dim, int i) {
  _grid grd = _DOM_GRD(*dom);

  if (i <= 3) {
    return 0;
  }
  else if (i >= (_GRD_SIZE(*grd, dim) - 1) * 4) {
    return _GRD_SIZE(*grd, dim) - 1;
  }
  else {
    return i / 4;
  }
}

