/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* to get definition of NULL */

#include "md_zinc.h"
#include "priv_access.h"
#include "region.h"
#include "zc_proto.h"
#include "zerror.h"
#include "zlib.h"


static void _SetupDynamicRegionNoDirections(_region_fnc DR) {
  int numdims;
  int i;
  int flooddim = 0;
  _domain dom = _REG_DOM(DR);

  numdims = _NUMDIMS(DR);
  for (i=0;i<numdims;i++) {
    if (_FLOODED_DIM(DR,i)) {
      _REG_LO(DR,i) = _DOM_LO(*dom, i);
      _REG_HI(DR,i) = _DOM_HI(*dom, i);
      _REG_ALIGN(DR,i) = 0;
      _REG_STRIDE(DR,i) = 1;
      flooddim = 1;
    }
  }

  _SPS_REGION(DR) = 0;
  _SetupRegion(DR);
  
  if (flooddim) {
    for (i=0;i<numdims;i++) {
      if (_FLOODED_DIM(DR,i)) {
	_REG_MYLO(DR,i) = _DOM_MYLO(*dom, i);
	_REG_MYHI(DR,i) = _DOM_MYLO(*dom, i);
      }
    }
  }
}


void _SetupDynamicRegion(_region_fnc DR) {
  _SetupDynamicRegionNoDirections(DR);
}


void _SetupWrapReflect(_region_fnc reg,_region basereg,_direction dir) {
  int sendwrap;
  int recvwrap;
  int srcproc = _INDEX;
  int dstproc = _INDEX;
  int i;
  int numdims;
  int dstloc[_MAXRANK];
  int srcloc[_MAXRANK];
  _domain dom = _REG_DOM(reg);
  _grid grd = _DOM_GRD(*dom);

  numdims = _NUMDIMS(reg);
  for (i=0; i<numdims; i++) {
    if (dir[i] < 0) {
      dstloc[i] = _MIN_PROC(reg,i);
      srcloc[i] = _MAX_PROC(basereg,i);
    } else if (dir[i] == 0) {
      dstloc[i] = _GRD_LOC(*grd, i);
      srcloc[i] = _GRD_LOC(*grd, i);
    } else {
      dstloc[i] = _MAX_PROC(reg,i);
      srcloc[i] = _MIN_PROC(basereg,i);
    }
  }
  dstproc = _GRD_TO_PROC(grd, numdims, dstloc);
  srcproc = _GRD_TO_PROC(grd, numdims, srcloc);

  sendwrap = (_INDEX == srcproc);
  recvwrap = (_INDEX == dstproc);

  /* This means "is it local for everyone?" not just "...for me?" */
  _WRAP_LOCAL(reg) = (dstproc == srcproc);

  if (sendwrap) {
    _I_SEND_WRAP(reg) = dstproc;
  } else {
    _I_SEND_WRAP(reg) = _NO_PROCESSOR;
  }
  if (recvwrap) {
    _I_RECV_WRAP(reg) = srcproc;
  } else {
    _I_RECV_WRAP(reg) = _NO_PROCESSOR;
  }
  for (i=0;i<numdims;i++) {
    if (dir[i] == 0) {
      _WRAP_SEND_LO(reg,i) = _REG_MYLO(basereg,i);
      _WRAP_SEND_HI(reg,i) = _REG_MYHI(basereg,i);
      _REFLECT_SRC_LO(reg,i) = _REG_MYLO(basereg,i);
      _REFLECT_SRC_HI(reg,i) = _REG_MYHI(basereg,i);
      _REFLECT_SRC_STEP(reg,i) = 1;
    } else if (dir[i] < 0) {
      _WRAP_SEND_LO(reg,i) = _SNAP_UP(_REG_MYHI(basereg,i)+dir[i]+1,basereg,i);
      _WRAP_SEND_HI(reg,i) = _SNAP_DN(_REG_MYHI(basereg,i),basereg,i);
      _REFLECT_SRC_LO(reg,i) = _REG_MYLO(basereg,i)- dir[i] - 1;
      _REFLECT_SRC_HI(reg,i) = _REG_MYLO(basereg,i);
      _REFLECT_SRC_STEP(reg,i) = -1;
    } else {
      _WRAP_SEND_LO(reg,i) = _SNAP_UP(_REG_MYLO(basereg,i),basereg,i);
      _WRAP_SEND_HI(reg,i) = _SNAP_DN(_REG_MYLO(basereg,i)+dir[i]-1,basereg,i);
      _REFLECT_SRC_LO(reg,i) = _REG_MYHI(basereg,i);
      _REFLECT_SRC_HI(reg,i) = _REG_MYHI(basereg,i) - dir[i] + 1;
      _REFLECT_SRC_STEP(reg,i) = -1;
    }
  }
}


void _AdjustRegionWithOfGlob(_region_fnc DR,_region prevR,_direction dir) {
  int numdims;
  int i;
  _PRIV_TID_DECL;

  _REG_DOM(DR) = _REG_DOM(prevR);
  numdims = _NUMDIMS(prevR);
  _NUMDIMS(DR) = numdims;

  for (i=0;i<numdims;i++) {
    if (dir[i] == 0) {
      _REG_LO(DR,i) = _REG_LO(prevR,i);
      _REG_HI(DR,i) = _REG_HI(prevR,i);
      _FLOODED_DIM(DR,i) = _FLOODED_DIM(prevR,i);
    } else {
      if (dir[i] < 0) {
	if (_SPS_REGION(prevR)) {
	  /* of regions for sparse regions are degenerate */
	  _REG_LO(DR,i) = _REG_LO(prevR,i);
	  _REG_HI(DR,i) = _REG_LO(prevR,i)-1;
	} else {
	  _REG_LO(DR,i) = _REG_LO(prevR,i)+dir[i];
	  _REG_HI(DR,i) = _REG_LO(prevR,i)-1;
	}
      } else if (dir[i] > 0) {
	if (_SPS_REGION(prevR)) {
	  /* of regions for sparse regions are degenerate */
	  _REG_LO(DR,i) = _REG_HI(prevR,i)+1;
	  _REG_HI(DR,i) = _REG_HI(prevR,i);
	} else {
	  _REG_LO(DR,i) = _REG_HI(prevR,i)+1;
	  _REG_HI(DR,i) = _REG_HI(prevR,i)+dir[i];
	}
      }
      if (_FLOODED_DIM(prevR,i)) {
	_RT_ALL_FATAL0("Attempt to make an of region from a floodable region\n"
		       "in a floodable dimension.");
      }
      _FLOODED_DIM(DR,i) = 0;
    }
    _REG_ALIGN(DR,i) = _REG_ALIGN(prevR,i);
    _REG_STRIDE(DR,i) = _REG_STRIDE(prevR,i);
    _REG_GLOB_LO(DR,i) = _SNAP_UP(_REG_LO(DR,i),DR,i);
    _REG_GLOB_HI(DR,i) = _SNAP_DN(_REG_HI(DR,i),DR,i);
  }
}


void _AdjustRegionWithOfLoc(_region_fnc DR, _region prevR, _direction dir) {
  int sparse;

  sparse = _SPS_REGION(prevR);
  _SPS_REGION(DR) = sparse;
  _SetupRegion(DR);
  if (sparse) {
    _SPS_INHERIT_STRUCTURE(DR,prevR,_NUMDIMS(DR));
  }
  _SetupWrapReflect(DR,prevR,dir);
}


void _AdjustRegionWithOf(_region_fnc DR,_region prevR,_direction dir) {
  _AdjustRegionWithOfGlob(DR,prevR,dir);
  _AdjustRegionWithOfLoc(DR,prevR,dir);
}


void _AdjustRegionWithAtGlob(_region_fnc DR,_region prevR,_direction dir) {
  int i;
  
  _REG_DOM(DR) = _REG_DOM(prevR);
  _NUMDIMS(DR) = _NUMDIMS(prevR);
  
  for (i=0;i<_NUMDIMS(DR);i++) {
    _REG_LO(DR,i) = _REG_LO(prevR,i) + dir[i];
    _REG_HI(DR,i) = _REG_HI(prevR,i) + dir[i];
    _REG_STRIDE(DR,i) = _REG_STRIDE(prevR,i);
    _REG_ALIGN(DR,i) = _SNAP_POSNEG(_REG_ALIGN(prevR,i)+dir[i],_REG_STRIDE(DR,i));
    _FLOODED_DIM(DR,i) = _FLOODED_DIM(prevR,i);
    _REG_GLOB_LO(DR,i) = _SNAP_UP(_REG_LO(DR,i),DR,i);
    _REG_GLOB_HI(DR,i) = _SNAP_DN(_REG_HI(DR,i),DR,i);
  }
}


void _AdjustRegionWithAtLoc(_region_fnc DR,_region prevR,_direction dir) {
  int sparse;

  sparse = _SPS_REGION(prevR);
  _SPS_REGION(DR) = sparse;
  _SetupRegion(DR);
  if (sparse) {
    _SpsCopyShiftStructure(DR,prevR,dir);
  }
}


void _AdjustRegionWithAt(_region_fnc DR,_region prevR,_direction dir) {
  _AdjustRegionWithAtGlob(DR,prevR,dir);
  _AdjustRegionWithAtLoc(DR,prevR,dir);
}


void _AdjustRegionWithInGlob(_region_fnc DR,_region prevR,_direction dir) {
  int i;

  _REG_DOM(DR) = _REG_DOM(prevR);
  _NUMDIMS(DR) = _NUMDIMS(prevR);

  for (i=0;i<_NUMDIMS(DR);i++) {
    if (dir[i] == 0) {
      _REG_LO(DR,i) = _REG_LO(prevR,i);
      _REG_HI(DR,i) = _REG_HI(prevR,i);
    } else if (dir[i] < 0) {
      _REG_LO(DR,i) = _REG_LO(prevR,i);
      _REG_HI(DR,i) = _REG_LO(prevR,i)-(dir[i]+1);
    } else {
      _REG_LO(DR,i) = _REG_HI(prevR,i)-(dir[i]-1);
      _REG_HI(DR,i) = _REG_HI(prevR,i);
    }
    _REG_ALIGN(DR,i) = _REG_ALIGN(prevR,i);
    _REG_STRIDE(DR,i) = _REG_STRIDE(prevR,i);
    _FLOODED_DIM(DR,i) = _FLOODED_DIM(prevR,i);
    _REG_GLOB_LO(DR,i) = _SNAP_UP(_REG_LO(DR,i),DR,i);
    _REG_GLOB_HI(DR,i) = _SNAP_DN(_REG_HI(DR,i),DR,i);
  }
}


void _AdjustRegionWithInLoc(_region_fnc DR,_region prevR,_direction dir) {
  int sparse;

  sparse = _SPS_REGION(prevR);
  _SPS_REGION(DR) = sparse;
  _SetupRegion(DR);
  if (sparse) {
    _SPS_INHERIT_STRUCTURE(DR,prevR,_NUMDIMS(DR));
    _SPS_NUMVALS(DR) = _SPS_NUMVALS(prevR);  /* conservative */
    _SPS_NUMNODES(DR) = _SPS_NUMNODES(prevR); /* ditto */
  }
}


void _AdjustRegionWithIn(_region_fnc DR,_region prevR,_direction dir) {
  _AdjustRegionWithInGlob(DR,prevR,dir);
  _AdjustRegionWithInLoc(DR,prevR,dir);
}


void _AdjustRegionWithByGlob(_region_fnc DR,_region prevR,_direction dir) {
  int i;
  int alignfrom;

  _REG_DOM(DR) = _REG_DOM(prevR);
  _NUMDIMS(DR) = _NUMDIMS(prevR);

  for (i=0;i<_NUMDIMS(DR);i++) {
    _REG_LO(DR,i) = _REG_LO(prevR,i);
    _REG_HI(DR,i) = _REG_HI(prevR,i);
    if (dir[i] < 0) {
      _REG_STRIDE(DR,i) = -_REG_STRIDE(prevR,i)*dir[i];
      alignfrom = _REG_GLOB_HI(prevR,i);
    } else {  /* maybe 0 should cause a flood dimension? */
      _REG_STRIDE(DR,i) = _REG_STRIDE(prevR,i)*dir[i];
      alignfrom = _REG_GLOB_LO(prevR,i);
    }
    _FLOODED_DIM(DR,i) = _FLOODED_DIM(prevR,i);
    if (_FLOODED_DIM(DR,i)) {
      _REG_STRIDE(DR,i) = 1;
    }
    _REG_ALIGN(DR,i) = _SNAP_LO_TO_ALIGN(alignfrom,_REG_STRIDE(DR,i));
    _REG_GLOB_LO(DR,i) = _SNAP_UP(_REG_LO(DR,i),DR,i);
    _REG_GLOB_HI(DR,i) = _SNAP_DN(_REG_HI(DR,i),DR,i);
  }
}


void _AdjustRegionWithByLoc(_region_fnc DR, _region prevR, _direction dir) {
  _SPS_REGION(DR) = _SPS_REGION(prevR);
  _SetupRegion(DR);
}


void _AdjustRegionWithBy(_region_fnc DR,_region prevR,_direction dir) {
  _AdjustRegionWithByGlob(DR,prevR,dir);
  _AdjustRegionWithByLoc(DR,prevR,dir);
}


void _SetupDynamicOfRegion(_region_fnc DR,_region_fnc base,_direction dir) {
  _SetupDynamicRegionNoDirections(base);
  _AdjustRegionWithOf(DR,base,dir);
}


void _SetupDynamicInRegion(_region_fnc DR,_region_fnc base,_direction dir) {
  _SetupDynamicRegionNoDirections(base);
  _AdjustRegionWithIn(DR,base,dir);
}


void _SetupDynamicAtRegion(_region_fnc DR,_region_fnc base,_direction dir) {
  _SetupDynamicRegionNoDirections(base);
  _AdjustRegionWithAt(DR,base,dir);
}


void _SetupDynamicByRegion(_region_fnc DR,_region_fnc base,_direction dir) {
  _SetupDynamicRegionNoDirections(base);
  _AdjustRegionWithBy(DR,base,dir);
}
