/******************************************************************************

            Copyright (c) 1996 - 2004  --  University of Washington

******************************************************************************/

#include "stand_inc.h"

/*
 *
 */
static void _ZPL_ALLOC_ARRAY_SPARSE(_array_fnc arr,
				    _region reg,
				    int elemsize,
				    int* up,
				    int* down,
				    int wrap,
				    char*(*fn)(_array, _vector)) {
  int lo[_MAXRANK], hi[_MAXRANK];
  int rank, i;

  rank = _NUMDIMS(reg);
  for (i = 0; i < rank; i++) {
    lo[i] = _REG_MYLO(reg, i);
    hi[i] = _REG_MYHI(reg, i);
  }
  _ARR_NUMDIMS(arr) = rank;
  _ARR_ELEMSIZE(arr) = elemsize;
  _ARR_BLK(arr, rank-1) = elemsize;
  _ARR_DATA_SIZE(arr) = _ARR_BLK(arr, rank-1) * _SPS_NUMVALS(reg);
  for (i = 0; i < rank; i++) {
    if (_FLOODED_DIM(reg, i) || (lo[i] == hi[i])) {
      _ARR_BLK(arr, i) = 0;
    }
  }
  _ARR_DATA(arr) = (char *)_zmalloc(_ARR_DATA_SIZE(arr),"sparse array data");
  _ARR_ORIGIN(arr) = _ARR_DATA(arr);
  _ARR_SPS_ZERO_IDENT(arr);
  _ARR_TYPE(arr)="whichever";
  _ARR_SET_ACCESSOR(arr, fn);
  _ARR_W(arr)=NULL;
  _ARR_R(arr)=NULL;
}

/*
 * up and down are for fluff
 */
void _ZPL_ALLOC_ARRAY(_array_fnc arr,
		      _region_fnc reg,
		      int elemsize,
		      int* up,
		      int* down,
		      int wrap,
		      int sparse,
		      char*(*fn)(_array, _vector),
		      char* label,
		      int exists,  /* use realloc instead of alloc */
		      int dnr) /* do not register */
{
  int lo[_MAXRANK], hi[_MAXRANK];
  int alo, ahi;
  int rank, i;
  int degenerate;
  int fluffinteresting;

  _ARR_LABEL(arr) = label;

  _ARR_DECL_REG(arr) = reg;
  if (!dnr) _REG_REGISTER(reg, arr);

  if (sparse) {
    _ZPL_ALLOC_ARRAY_SPARSE(arr, reg, elemsize, up, down, wrap, fn);
    return;
  }

  rank = _NUMDIMS(reg);

  fluffinteresting = 0;
  if (up != NULL) {
    for (i=0; i<rank; i++) {
      if (up[i] != 0 || down[i] != 0) {
	fluffinteresting = 1;
      }
    }
  }

  if (fluffinteresting == 0) {
    /* if there is no fluff, we simply use the region bounds */
    for (i=0; i<rank; i++) {
      lo[i] = _REG_MYLO(reg, i);
      hi[i] = _REG_MYHI(reg, i);
    }
  } else {
    /* if there is fluff, then we want to expand our domain bounds by
       the fluff amounts (making sure not to wrap past INT_MIN/MAX)
       and intersect it with the array's declared region to see what
       bounds we should allocate.  It should be noted that even if a
       processor owns no indices of a region, it may need to allocate
       fluff for that region.

       Wrap-@'s present another weird case.  In this case, we spoof
       the region's bounds to make it appear larger than it truly is.
    */

    _reg_info fakereginfo;
    _region_fnc fakereg = &fakereginfo;
    _domain dom = _REG_DOM(reg);
    _grid grd = _DOM_GRD(*dom);
    int domlo;
    int domhi;

    for (i=0; i<rank; i++) {
      _REG_LO(fakereg, i) = _REG_LO(reg, i);
      _REG_HI(fakereg, i) = _REG_HI(reg, i);
      _REG_STRIDE(fakereg, i) = _REG_STRIDE(reg, i);
      _REG_ALIGN(fakereg, i) = _REG_ALIGN(reg, i);

      if (wrap && (up != NULL)) {
	_REG_LO(fakereg, i) -= down[i];
	_REG_HI(fakereg, i) += up[i];
      }
      _REG_GLOB_LO(fakereg, i) = _SNAP_UP(_REG_LO(fakereg, i), reg, i);
      _REG_GLOB_HI(fakereg, i) = _SNAP_DN(_REG_HI(fakereg, i), reg, i);

      domlo = _DOM_MYLO(*dom, i);
      domhi = _DOM_MYHI(*dom, i);
      if (domlo - down[i] > domlo) {
	domlo = INT_MIN;
      } else {
	domlo = domlo - down[i];
      }
      if (domhi + up[i] < domhi) {
	domhi = INT_MAX;
      } else {
	domhi = domhi + up[i];
      }

#ifdef DEBUG_FLUFF_DISTRIB
      printf("[%d] sending in %d..%d and %d..%d\n", _INDEX,
	     _REG_GLOB_LO(fakereg, i), _REG_GLOB_HI(fakereg, i),
	     domlo, domhi);
#endif
      
      _SetMyBlock(i, fakereg, domlo, domhi, grd);
      lo[i] = _REG_MYLO(fakereg, i);
      hi[i] = _REG_MYHI(fakereg, i);
#ifdef DEBUG_FLUFF_DISTRIB
      printf("[%d] I own %d..%d by %d of %d\n", _INDEX, lo[i], hi[i], 
	     _REG_STRIDE(reg, i), i);
#endif
    }
  }

  degenerate = 0;
  for (i = 0; i < rank; i++) {
    if (lo[i] > hi[i]) {
      degenerate = 1;
    }
  }
  _ARR_NUMDIMS(arr) = rank;
  _ARR_ELEMSIZE(arr) = elemsize;
  if (!degenerate) {
    for (i = 0; i < rank; i++) {
      _ARR_STR(arr, i) = _REG_STRIDE(reg, i);
      _ARR_OFF(arr, i) = lo[i];
    }
    _ARR_BLK(arr, rank-1) = elemsize;
    for (i = rank - 1; i > 0; i--) {
      _ARR_BLK(arr, i-1) = _ARR_BLK(arr, i) *
	                   (((hi[i] - lo[i])/_ARR_STR(arr, i))+1);
    }
  } else {
    for (i = 0; i < rank; i++) {
      _ARR_STR(arr, i) = 1;
      _ARR_OFF(arr, i) = 0;
      _ARR_BLK(arr, i) = 0;
    }
  }
  _ARR_DATA_SIZE(arr) = _ARR_BLK(arr, 0) * ((hi[0] - lo[0])/_ARR_STR(arr, 0) + 1);
  for (i = 0; i < rank; i++) {
    if (_FLOODED_DIM(reg, i) || (lo[i] == hi[i])) {
      _ARR_BLK(arr, i) = 0;
    }
  }
  if (!exists) {
    _ARR_DATA(arr) = (char *)_zmalloc(_ARR_DATA_SIZE(arr),"array data");
  }
  else {
    _ARR_DATA(arr) = (char *)_zrealloc(_ARR_DATA(arr), _ARR_DATA_SIZE(arr),"array data");
  }
  switch(rank) {
  case 1:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 1);
    break;
  case 2:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 2);
    break;
  case 3:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 3);
    break;
  case 4:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 4);
    break;
  case 5:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 5);
    break;
  case 6:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 6);
    break;
  default:
    _ARR_ORIGIN(arr) = _ARR_COMP_ORIG(arr, 1);
    break;
  }
  _ARR_TYPE(arr)="whichever";
  _ARR_SET_ACCESSOR(arr, fn);
  _ARR_W(arr)=NULL;
  _ARR_R(arr)=NULL;
}

void _ZPL_FREE_ARRAY(_array_fnc arr, int dnr) {
  if (!dnr) _REG_UNREGISTER(_ARR_REG(arr), arr);
  _zfree(_ARR_DATA(arr), "array data");
}

void _BOUND_CHECK(_region reg, _array arr, int lineno) {
  _region areg = _ARR_DECL_REG(arr);
  int i;

  for (i = 0; i < _NUMDIMS(reg); i++) {
    if (_REG_GLOB_LO(reg, i) < _REG_GLOB_LO(areg, i) ||
	_REG_GLOB_HI(reg, i) > _REG_GLOB_HI(areg, i)) {
      _SCALAR_FPRINTF1(zout, "Access out of bounds at line %d", lineno);
      _SCALAR_LINEFEED(zout);
      _ZPL_halt(lineno);
    }
  }
}
